@@ -23,9 +23,16 @@
 #include <stdlib.h>
 #include <string.h>
 #include <popt.h>
+#include <sys/time.h> /* TODO: replace timeofday with g_timers */
 #include <gdk/gdkwindow.h>
 #include <gtk/gtk.h>
+#ifndef USE_GDKPIXBUF_RENDER_TO_DRAWABLE
+#include <gdk/gdkpixbuf.h>
+#else
+#include <gdk/gdk.h>
+#endif
 #include <gdk/gdkx.h>
+#include <gdk/gdkrgb.h>
 #include <libbonobo.h>
 #include <X11/Xlib.h>
 #include <X11/Xutil.h>
@@ -66,25 +73,17 @@ typedef enum {
 } ZoomRegionPixmapCreationError;
 
 /*
- * When running the magnifier with the panning test (--timing-pan-rate), the following
- * output is displayed:
+ * When running the magnifier in timing test mode (--timing-pan-rate and/or --timing-iterations),
+ * then the following output is displayed:
  *
- * ### Updates summary:
- * 
+ *   Pan Rate                 = 200
+ *   Frames Processed         = 1225
+ *   Width/Height/Depth       = 640/1024/8
+ *   Zoom Factor (x/y)        = 3.000000/3.000000
  *   Update Duration          = (avg. 0.010232) (max. 0.002685) (tot. 1.084589) seconds
  *     Update Pixels          = (avg. 3046) pixels/frame
  *     Update Rate            = (avg. 97.732909) (max. 372.439485) updates/second
  *     Net Update Rate        = (avg. 0.297724) (max. 0.457741) Mpex/second
- * 
- * ### Updates finished, starting panning test
- * 
- *   Pan Rate                 = 200
- *
- * Final Report:
- *
- *   Frames Processed         = 1225
- *   Width/Height/Depth       = 640/1024/8
- *   Zoom Factor (x/y)        = 3.000000/3.000000
  *   Pan Latency              = (avg. 0.008229) (max. 0.562326) seconds
  *   Total Frame Duration     = (avg. 0.008363) (max. 0.562485) (tot. 10.235981) seconds
  *   Frame Rate               = (avg. 119.578189) (max. 4854.369141) frames/second
@@ -93,31 +92,33 @@ typedef enum {
  *   Scroll Rate (x)          = (avg. 0.000000) lines/second
  *   Scroll Rate (y)          = (avg. 600.235657) lines/second
  *   Net Render Rate          = (avg. 8.685322) (max. 3181.359360) Mpex/second
- * 
+ *
  * Pan Rate             - Displays the pan rate specified on the command line.
+ *                        Only displayed if --timing-pan-rate is used.
  * Frames Processed     - Number of frames displayed
  * Width/Height/Depth   - The width/height/depth of magnifier display.
  * Zoom Factor (x/y)    - The zoom factor used.
+ * Update Duration      - Average and total time spent updating new sections of the screen.
+ *   Update Pixels      - Number of pixels updated
+ *   Update Rate        - Number of updates possible per second.
+ *   Net Update Rate    - Number of Megapixels updated per second.
  * Pan Latency          - Time spent from the beginning of the setROI request until GTK+
- *                        has finished processing the update (in GDK's 
+ *                        has finished processing the update (in GDK's
  *                        gdk_window_update_idle idle handler.
  * Total Frame Duration - Average and total time processing frames.
  * Frame Rate           - Number of frames per second.
  * Scroll Delta         - Average and total number of lines scrolled.
- * Scroll Rate          - Number of lines scrolled per second.
+ * Scroll Rate          - Number of lines scrolled per second.  Note that this value
+ *                        will be equal to the --timing-pan-rate value times the
+ *                        y zoom factor.
  * Net Render Rate      - Number of Megapixels processed per second.
- * 
- * Note that updates do not happen for some frames, such as when panning.  When updates
- * do happen, the following lines are included:
- *
- * Update Duration   - Average and total time spent updating new sections of the screen.
- *   Update Pixels   - Number of pixels updated
- *   Update Rate     - Number of updates possible per second.
- *   Net Update Rate - Number of Megapixels updated per second.
  *
  * When running with the --timing-output argument, similar information is displayed for
- * each frame plus timing information for the frame.
- */ 
+ * each frame plus timing information for the frame.  Update information is not
+ * displayed for frames where updates do not occur.  Pan Rate, Frames Processed,
+ * and Width/Height/Depth is not displayed for each frame, only when displaying
+ * summary information.
+ */
 
 typedef struct timing {
 	struct timeval process_start;
@@ -154,6 +155,7 @@ static float cps_max           = 0;
 static float nrr_max           = 0;
 static float update_nrr_max    = 0;
 static gboolean	reset_timing   = FALSE;
+static GdkGC *default_gc       = NULL;
 static gboolean timing_test    = FALSE;
 
 static guint pending_idle_handler = 0;
@@ -183,6 +185,12 @@ static int  zoom_region_update_pointer_timeout (gpointer data);
 void
 reset_timing_stats()
 {
+	timing_scale_max               = 0;
+	timing_idle_max                = 0;
+	timing_frame_max               = 0;
+	cps_max                        = 0;
+	nrr_max                        = 0;
+	update_nrr_max                 = 0;
 	mag_timing.num_scale_samples   = 0;
 	mag_timing.num_idle_samples    = 0;
 	mag_timing.num_frame_samples   = 0;
@@ -597,13 +605,6 @@ zoom_region_clip_to_exposed_bounds (ZoomRegion *zoom_region,
 	return _rectangle_clip_to_bounds (area, &zoom_region->priv->exposed_bounds);
 }
 
-static GdkRectangle
-zoom_region_clip_to_target (ZoomRegion *zoom_region,
-			    GdkRectangle area)
-{
-	return _rectangle_clip_to_bounds (area, &zoom_region->priv->source_area);
-}
-
 static GdkRectangle
 zoom_region_clip_to_viewport (ZoomRegion *zoom_region,
 			      GdkRectangle area)
@@ -612,6 +613,13 @@ zoom_region_clip_to_viewport (ZoomRegion *zoom_region,
 }
 #endif
 
+static const GdkRectangle
+zoom_region_clip_to_target (ZoomRegion *zoom_region,
+			    GdkRectangle area)
+{
+	return _rectangle_clip_to_bounds (area, &zoom_region->priv->source_area);
+}
+
 static GdkRectangle
 zoom_region_clip_to_window (ZoomRegion *zoom_region,
 			    GdkRectangle area)
@@ -624,7 +632,7 @@ zoom_region_clip_to_window (ZoomRegion *zoom_region,
 	return _rectangle_clip_to_rectangle (area, window_rect);
 }
 
-static GdkRectangle
+static const GdkRectangle
 zoom_region_source_rect_from_view_bounds (ZoomRegion *zoom_region,
 					  const GNOME_Magnifier_RectBounds *view_bounds)
 {
@@ -756,39 +764,69 @@ static void
 zoom_region_paint_crosswire_cursor (ZoomRegion *zoom_region, GdkRectangle *clip_rect)
 {
 	Magnifier *magnifier = zoom_region->priv->parent;
-	GdkGC *gc;
-	GdkColor color = {0, 0, 0, 0};
-	int x_left_clip = 0, x_right_clip = 0, y_top_clip = 0, y_bottom_clip = 0;
-	int csize;
+	static GdkGC *crosswire_gc = NULL;
+	static GdkColormap *cmap;
+	static GdkColor last_color;
+	static gboolean last_color_init = !TRUE;
+	GdkGCValues values;
 	GdkRectangle rect;
 	GdkDrawable *cursor;
+	GdkColor color = {0, 0, 0, 0};
+	GdkColor tmp_color;
+	int x_left_clip = 0, x_right_clip = 0, y_top_clip = 0, y_bottom_clip = 0;
+	int csize;
 	
 	if (!(magnifier && zoom_region->priv->w &&
 	      zoom_region->priv->w->window &&
 	      magnifier->crosswire_size > 0)) return;
 
+	if (crosswire_gc == NULL) 
+	{
+		crosswire_gc = gdk_gc_new (zoom_region->priv->w->window);
+		cmap = gdk_gc_get_colormap(crosswire_gc);
+	}
 
-	gc = gdk_gc_new (zoom_region->priv->w->window);
 	if (magnifier->crosswire_color == 0)
 	{
 		color.red = 0xFF;
 		color.blue = 0xFF;
 		color.green = 0xFF;
-		gdk_gc_set_function (gc, GDK_INVERT);
+		values.function = GDK_INVERT;
 	}
 	else
 	{
 		color.red = (magnifier->crosswire_color & 0xFF0000) >> 8;
 		color.green = (magnifier->crosswire_color & 0xFF00);
 		color.blue = (magnifier->crosswire_color & 0xFF) << 8;
-		gdk_gc_set_function (gc, GDK_COPY);
+		values.function = GDK_COPY;
+	}
+
+	values.foreground = color;
+
+	/* Only reset colors if they have changed */
+    if (!last_color_init || color.red != last_color.red ||
+	    color.blue != last_color.blue || color.green != last_color.green)
+	{
+		if (cmap)
+		{
+			gdk_rgb_find_color (cmap, &(values.foreground));
+			gdk_gc_set_values(crosswire_gc, &values, GDK_GC_FUNCTION | GDK_GC_FOREGROUND);
+		}
+		else
+		{
+			gdk_gc_set_values(crosswire_gc, &values, GDK_GC_FUNCTION);
+		}
+
+		last_color.red   = color.red;
+		last_color.blue  = color.blue;
+		last_color.green = color.green;
+		last_color_init  = TRUE;
 	}
-	gdk_gc_set_rgb_fg_color (gc, &color);
 
 	rect.x = zoom_region->priv->last_cursor_pos.x;
 	rect.y = zoom_region->priv->last_cursor_pos.y;
 	rect = zoom_region_view_rect_from_source_rect (zoom_region, rect);
-	if (clip_rect) gdk_gc_set_clip_rectangle (gc, clip_rect);
+	if (clip_rect) gdk_gc_set_clip_rectangle (crosswire_gc, clip_rect);
 
 	if (magnifier->crosswire_clip &&
 	    (cursor = magnifier_get_cursor (magnifier)))
@@ -809,31 +847,31 @@ zoom_region_paint_crosswire_cursor (ZoomRegion *zoom_region, GdkRectangle *clip_
 	{
 		if (magnifier->crosswire_clip)
 		{
-			gdk_draw_line (zoom_region->priv->w->window, gc, rect.x, 0,
+			gdk_draw_line (zoom_region->priv->w->window, crosswire_gc, rect.x, 0,
 				       rect.x, y_top_clip);
-			gdk_draw_line (zoom_region->priv->w->window, gc, 0, rect.y,
+			gdk_draw_line (zoom_region->priv->w->window, crosswire_gc, 0, rect.y,
 				       x_left_clip, rect.y);
 		}
-		gdk_draw_line (zoom_region->priv->w->window, gc, rect.x,
+		gdk_draw_line (zoom_region->priv->w->window, crosswire_gc, rect.x,
 			       y_bottom_clip, rect.x, 4096);
-		gdk_draw_line (zoom_region->priv->w->window, gc, x_right_clip,
+		gdk_draw_line (zoom_region->priv->w->window, crosswire_gc, x_right_clip,
 			       rect.y, 4096, rect.y);
 	}
 	else
 	{
 		if (magnifier->crosswire_clip )
 		{
-			gdk_draw_rectangle (zoom_region->priv->w->window, gc, TRUE,
+			gdk_draw_rectangle (zoom_region->priv->w->window, crosswire_gc, TRUE,
 					    rect.x - magnifier->crosswire_size / 2,
 					    0, magnifier->crosswire_size, y_top_clip);
-			gdk_draw_rectangle (zoom_region->priv->w->window, gc, TRUE, 0,
+			gdk_draw_rectangle (zoom_region->priv->w->window, crosswire_gc, TRUE, 0,
 					    rect.y - magnifier->crosswire_size / 2,
 					    x_left_clip, magnifier->crosswire_size);
 		}
-		gdk_draw_rectangle (zoom_region->priv->w->window, gc, TRUE,
+		gdk_draw_rectangle (zoom_region->priv->w->window, crosswire_gc, TRUE,
 				    rect.x - magnifier->crosswire_size / 2,
 				    y_bottom_clip, magnifier->crosswire_size, 4096);
-		gdk_draw_rectangle (zoom_region->priv->w->window, gc, TRUE, x_right_clip,
+		gdk_draw_rectangle (zoom_region->priv->w->window, crosswire_gc, TRUE, x_right_clip,
 				    rect.y - magnifier->crosswire_size / 2,
 				    4096, magnifier->crosswire_size);
 	}
@@ -849,6 +887,8 @@ static void
 zoom_region_paint_cursor (ZoomRegion *zoom_region,
 			  GdkRectangle *clip_rect)
 {
+	static GdkGC *paint_cursor_gc = NULL;
+	GdkGCValues values;
 	GdkRectangle rect, intersct;
 	GdkRectangle fullscreen;
 	Magnifier *magnifier = zoom_region->priv->parent;
@@ -862,7 +902,7 @@ zoom_region_paint_cursor (ZoomRegion *zoom_region,
 	if (gdk_rectangle_intersect (clip_rect, &rect, &intersct))
 	{
 		int width = 0, height = 0;
-		GdkGC *gc = gdk_gc_new (zoom_region->priv->w->window);
+		
 		GdkDrawable *cursor = magnifier_get_cursor (magnifier);
 		if (!cursor)
 			return;
@@ -873,7 +913,9 @@ zoom_region_paint_cursor (ZoomRegion *zoom_region,
 		}
 		if (rect.width != width || rect.height != height)
 		{
-			g_object_unref (zoom_region->priv->cursor_backing_pixels);
+			if (zoom_region->priv->cursor_backing_pixels) {
+				g_object_unref (zoom_region->priv->cursor_backing_pixels);
+			}
 			zoom_region->priv->cursor_backing_pixels =
 				gdk_pixmap_new (zoom_region->priv->w->window,
 						rect.width,
@@ -881,28 +923,40 @@ zoom_region_paint_cursor (ZoomRegion *zoom_region,
 						-1);
 		}
 		if (zoom_region->priv->w->window != NULL)
-		  gdk_draw_drawable (zoom_region->priv->cursor_backing_pixels,
-				     gc,
+		{
+			if (default_gc == NULL)
+				default_gc = gdk_gc_new(zoom_region->priv->w->window);
+			gdk_draw_drawable (zoom_region->priv->cursor_backing_pixels,
+				     default_gc,
 				     zoom_region->priv->w->window,
 				     rect.x,
 				     rect.y,
 				     0, 0,
 				     rect.width,
 				     rect.height);
+		}
 		DEBUG_RECT ("painting", rect);
-		gdk_gc_set_clip_rectangle (gc, clip_rect);
-		gdk_gc_set_clip_mask (gc, magnifier->priv->cursor_mask);
-		gdk_gc_set_clip_origin (gc, rect.x, rect.y);
 		if (cursor)
+		{
+			if (paint_cursor_gc == NULL)
+				paint_cursor_gc = gdk_gc_new (zoom_region->priv->w->window);
+
+			gdk_gc_set_clip_rectangle (paint_cursor_gc, clip_rect);
+			values.clip_x_origin = rect.x;
+			values.clip_y_origin = rect.y;
+			values.clip_mask = magnifier->priv->cursor_mask;
+			gdk_gc_set_values(paint_cursor_gc, &values, GDK_GC_CLIP_X_ORIGIN |
+				GDK_GC_CLIP_Y_ORIGIN | GDK_GC_CLIP_MASK);
+
 			gdk_draw_drawable (zoom_region->priv->w->window,
-					   gc,
+					   paint_cursor_gc,
 					   cursor,
 					   0, 0,
 					   rect.x,
 					   rect.y,
 					   rect.width,
 					   rect.height);
-		g_object_unref (gc);
+		}
 	}
 }
 
@@ -949,6 +1003,7 @@ zoom_region_paint_border (ZoomRegion *zoom_region,
 			  GdkRectangle *area)
 {
 	GdkColor color;
+	/*
 	if (zoom_region->border_size > 0)
 		if (!zoom_region->priv->border_gc)
 		{
@@ -962,7 +1017,6 @@ zoom_region_paint_border (ZoomRegion *zoom_region,
 #endif
 			gdk_gc_set_foreground (zoom_region->priv->border_gc, &color);
 		}
-	/*
 		gdk_draw_rectangle (zoom_region->priv->w->window,
 				    zoom_region->priv->border_gc,
 				    TRUE,
@@ -977,13 +1031,13 @@ static void
 zoom_region_paint_pixmap (ZoomRegion *zoom_region,
 			  GdkRectangle *area)
 {
-	GdkGC *gc = NULL;
-	if (zoom_region->priv->w && zoom_region->priv->w->window)
-		gc = gdk_gc_new (GDK_DRAWABLE (zoom_region->priv->w->window));
-	if (gc && zoom_region->priv->pixmap)
+	if (default_gc == NULL)
+		default_gc = gdk_gc_new(zoom_region->priv->w->window);
+
+	if (zoom_region->priv->pixmap)
 	{
 		gdk_draw_drawable (zoom_region->priv->w->window,
-				   gc,
+				   default_gc,
 				   zoom_region->priv->pixmap,
 				   area->x + zoom_region->priv->exposed_bounds.x1,
 				   area->y + zoom_region->priv->exposed_bounds.y1,
@@ -992,7 +1046,6 @@ zoom_region_paint_pixmap (ZoomRegion *zoom_region,
 				   area->width,
 				   area->height);
 	}
-	g_object_unref (gc); /* OK if gc is NULL */
 }
 
 /**
@@ -1503,8 +1556,12 @@ zoom_region_get_source_subwindow (ZoomRegion *zoom_region,
 					gdk_image_put_pixel (test_image, i, j, i*j);
 
 			zoom_region->priv->source_drawable = gdk_pixmap_new (NULL, width, height, 24);
+
+			if (default_gc == NULL)
+				default_gc = gdk_gc_new(zoom_region->priv->w->window);
+
 			gdk_draw_image (zoom_region->priv->source_drawable,
-					gdk_gc_new (zoom_region->priv->source_drawable),
+					default_gc,
 					test_image,
 					0, 0,
 					0, 0,
@@ -1573,7 +1630,6 @@ zoom_region_update (ZoomRegion *zoom_region,
 		    const GdkRectangle update_rect)
 {
 	GdkPixbuf *subimage;
-	GdkGC *gc;
 	GdkRectangle source_rect;
 	DEBUG_RECT ("unclipped update rect", update_rect);
 	source_rect = zoom_region_clip_to_exposed_target (zoom_region, update_rect);
@@ -1604,11 +1660,26 @@ zoom_region_update (ZoomRegion *zoom_region,
 
 		zoom_region_post_process_pixbuf (zoom_region, subimage,
 						 zoom_region->priv->scaled_pixbuf);
-		gc = gdk_gc_new (zoom_region->priv->w->window);
+		if (default_gc == NULL)
+			default_gc = gdk_gc_new(zoom_region->priv->w->window);
 
+#ifndef USE_GDK_PIXBUF_RENDER_TO_DRAWABLE 
+		gdk_draw_pixbuf (zoom_region->priv->pixmap,
+				 default_gc,
+				 zoom_region->priv->scaled_pixbuf,
+				 0,
+				 0,
+				 paint_rect.x + zoom_region->priv->exposed_bounds.x1,
+				 paint_rect.y + zoom_region->priv->exposed_bounds.y1,
+				 paint_rect.width,
+				 paint_rect.height,
+				 GDK_RGB_DITHER_NONE,
+				 0,
+				 0);
+#else
 		gdk_pixbuf_render_to_drawable (zoom_region->priv->scaled_pixbuf,
 					       zoom_region->priv->pixmap,
-					       gc,
+					       default_gc,
 					       0,
 					       0,
 					       paint_rect.x + zoom_region->priv->exposed_bounds.x1,
@@ -1618,9 +1689,9 @@ zoom_region_update (ZoomRegion *zoom_region,
 					       GDK_RGB_DITHER_NONE,
 					       0,
 					       0);
+#endif
 		if (magnifier_error_check ())
 			g_warning ("Could not render scaled image to drawable; out of memory!\n");
-		g_object_unref (gc);
 		g_object_unref (subimage);
 		gdk_window_begin_paint_rect (zoom_region->priv->w->window, &paint_rect);
 		zoom_region_paint (zoom_region, &paint_rect);
@@ -1647,7 +1718,7 @@ zoom_region_update (ZoomRegion *zoom_region,
 					mag_timing.scale_val, (mag_timing.scale_total / 
 					mag_timing.num_scale_samples), timing_scale_max, mag_timing.scale_total);
 				fprintf(stderr, "    Update Pixels          = %ld (avg. %ld) pixels/frame\n",
-					source_rect.height * source_rect.width,
+					(long) source_rect.height * source_rect.width,
 					mag_timing.update_pixels_total / mag_timing.num_scale_samples);
 				fprintf(stderr, "    Update Rate            = (avg. %f) (max. %f) updates/second\n",
 					1.0/(mag_timing.scale_total / mag_timing.num_scale_samples), 1.0/(float)timing_scale_max);
@@ -1722,6 +1793,67 @@ zoom_region_process_updates (gpointer data)
 		return FALSE;
 }
 
+void
+timing_report(ZoomRegion *zoom_region)
+{
+	float frame_avg;
+	float x_scroll_incr, y_scroll_incr;
+	int width, height, x, y;
+
+	if (timing_test) {
+		width = (zoom_region->priv->exposed_viewport.x2 -
+			zoom_region->priv->exposed_viewport.x1) / zoom_region->xscale;
+		height = (zoom_region->priv->exposed_viewport.y2 -
+			zoom_region->priv->exposed_viewport.y1) / zoom_region->yscale;
+
+		frame_avg = mag_timing.frame_total / mag_timing.num_frame_samples;
+
+		x_scroll_incr = (float)mag_timing.dx_total / (float)mag_timing.num_line_samples;
+		y_scroll_incr = (float)mag_timing.dy_total / (float)mag_timing.num_line_samples;
+
+		gdk_drawable_get_size (GDK_DRAWABLE (zoom_region->priv->w->window),
+			&x, &y);
+
+		fprintf(stderr, "  Frames Processed         = %ld\n", 
+			mag_timing.num_frame_samples + 1);
+		fprintf(stderr, "  Width/Height/Depth       = %d/%d/%d\n", x, y,
+			gdk_drawable_get_depth (zoom_region->priv->w->window));
+		fprintf(stderr, "  Zoom Factor (x/y)        = %f/%f\n", zoom_region->xscale,
+			zoom_region->yscale);
+		if (mag_timing.num_scale_samples != 0) {
+			fprintf(stderr, "  Update Duration          = (avg. %f) (max. %f) (tot. %f) seconds\n",
+				(mag_timing.scale_total / mag_timing.num_scale_samples), timing_scale_max, mag_timing.scale_total);
+			fprintf(stderr, "    Update Pixels          = (avg. %ld) pixels/frame\n",
+				mag_timing.update_pixels_total / mag_timing.num_scale_samples);
+			fprintf(stderr, "    Update Rate            = (avg. %f) (max. %f) updates/second\n",
+				1.0/((float)mag_timing.scale_total / (float)mag_timing.num_scale_samples),
+				1.0/(float)timing_scale_max);
+			fprintf(stderr, "    Net Update Rate        = (avg. %f) (max. %f) Mpex/second\n",
+				((float)mag_timing.update_pixels_total / (float)mag_timing.scale_total) / 1000000.0,
+				update_nrr_max / 1000000.0);
+                }
+		fprintf(stderr, "  Pan Latency              = (avg. %f) (max. %f) seconds\n",
+			(mag_timing.idle_total / mag_timing.num_idle_samples), timing_idle_max);
+		fprintf(stderr, "  Total Frame Duration     = (avg. %f) (max. %f) (tot. %f) seconds\n",
+			frame_avg, timing_frame_max, mag_timing.frame_total);
+		fprintf(stderr, "  Frame Rate               = (avg. %f) (max. %f) frames/second\n",
+			1.0 / (mag_timing.frame_total / mag_timing.num_frame_samples), cps_max);
+		fprintf(stderr, "  Scroll Delta (x)         = (avg. %f) (tot. %d) lines\n",
+			x_scroll_incr, mag_timing.dx_total);
+		fprintf(stderr, "  Scroll Delta (y)         = (avg. %f) (tot. %d) lines\n",
+			y_scroll_incr, mag_timing.dy_total);
+		fprintf(stderr, "  Scroll Rate (x)          = (avg. %f) lines/second\n",
+			x_scroll_incr / frame_avg);
+		fprintf(stderr, "  Scroll Rate (y)          = (avg. %f) lines/second\n",
+			y_scroll_incr / frame_avg);
+
+		fprintf(stderr, "  Net Render Rate          = (avg. %f) (max. %f) Mpex/second\n\n",
+			(height * width *
+			((float)mag_timing.num_frame_samples / (float)mag_timing.frame_total)) / 1000000.0,
+			nrr_max / 1000000.0);
+	}
+}
+
 static void
 zoom_region_time_frame(ZoomRegion *zoom_region, Magnifier *magnifier)
 {
@@ -1775,16 +1907,7 @@ zoom_region_time_frame(ZoomRegion *zoom_region, Magnifier *magnifier)
 
 	if (reset_timing) {
 		fprintf(stderr, "\n### Updates summary:\n\n");
-		fprintf(stderr, "  Update Duration          = (avg. %f) (max. %f) (tot. %f) seconds\n",
-			(mag_timing.scale_total / mag_timing.num_scale_samples), timing_scale_max, mag_timing.scale_total);
-		fprintf(stderr, "    Update Pixels          = (avg. %ld) pixels/frame\n",
-			mag_timing.update_pixels_total / mag_timing.num_scale_samples);
-		fprintf(stderr, "    Update Rate            = (avg. %f) (max. %f) updates/second\n",
-			1.0/((float)mag_timing.scale_total / (float)mag_timing.num_scale_samples),
-			1.0/(float)timing_scale_max);
-		fprintf(stderr, "    Net Update Rate        = (avg. %f) (max. %f) Mpex/second\n",
-			((float)mag_timing.update_pixels_total / (float)mag_timing.scale_total) / 1000000.0,
-			update_nrr_max / 1000000.0);
+		timing_report (zoom_region);
 	    	fprintf(stderr, "\n### Updates finished, starting panning test\n");
 		reset_timing_stats();
 		reset_timing = FALSE;
@@ -1814,7 +1937,7 @@ exit_report(ZoomRegion *zoom_region)
 
 		fprintf(stderr, "\nFinal Report:\n\n");
 
-		fprintf(stderr, "  Frames Processed         = %d\n", 
+		fprintf(stderr, "  Frames Processed         = %ld\n", 
 			mag_timing.num_frame_samples + 1);
 		fprintf(stderr, "  Width/Height/Depth       = %d/%d/%d\n", x, y,
 			gdk_drawable_get_depth (zoom_region->priv->w->window));
@@ -1840,7 +1963,6 @@ exit_report(ZoomRegion *zoom_region)
 			((float)mag_timing.num_frame_samples / (float)mag_timing.frame_total)) / 1000000.0,
 			nrr_max / 1000000.0);
 	}
-	exit(0);
 }
 
 static void
@@ -1894,7 +2016,6 @@ zoom_region_align (ZoomRegion *zoom_region)
 		gettimeofday(&mag_timing.frame_start, NULL);
 
 		if (zoom_region->timing_output) {
-			Magnifier *magnifier = zoom_region->priv->parent;
 			gint x, y;
 
 			gdk_drawable_get_size (GDK_DRAWABLE (zoom_region->priv->w->window),
@@ -1903,7 +2024,7 @@ zoom_region_align (ZoomRegion *zoom_region)
 			fprintf(stderr, "\nTiming Information - ROI   = (%d, %d) (%d, %d):\n",
 				zoom_region->roi.x1, zoom_region->roi.y1, zoom_region->roi.x2,
 				zoom_region->roi.y2);
-			fprintf(stderr, "  Frame Number             = %d\n", 
+			fprintf(stderr, "  Frame Number             = %ld\n", 
 				mag_timing.num_frame_samples + 1);
 			fprintf(stderr, "  Width/Height/Depth       = %d/%d/%d\n", x, y,
 				gdk_drawable_get_depth (zoom_region->priv->w->window));
@@ -2186,8 +2307,6 @@ static int
 zoom_region_process_pending (gpointer data)
 {
 	ZoomRegion *zoom_region = (ZoomRegion *) data;
-	Magnifier *magnifier = zoom_region->priv->parent;
-
 	zoom_region_align (zoom_region);
 	return FALSE;
 }
@@ -2201,11 +2320,11 @@ zoom_region_pan_test (gpointer data)
 	GNOME_Magnifier_RectBounds roi;
 	CORBA_Environment ev;
 	static int counter = 0;
-	static gboolean last_update = !TRUE;
 	static gboolean finished_update = !TRUE;
+    static float last_pixels_at_speed = -1;
 	float pixels_at_speed;
 	float total_time;
-	int screen_height, height, width;
+	int screen_height, height;
 	int pixel_position;
 	int pixel_direction;
 
@@ -2229,6 +2348,10 @@ zoom_region_pan_test (gpointer data)
 	else
 		pixels_at_speed = 0.0;
 
+    /* Wait until it is actually necessary to update the screen */
+    if ((int)(last_pixels_at_speed) == (int)(pixels_at_speed))
+        return TRUE;
+
 	pixel_position = (int)(pixels_at_speed) % (screen_height - height);
 	counter = (int)(pixels_at_speed) / (screen_height - height);
 	pixel_direction = counter % 2;
@@ -2241,7 +2364,6 @@ zoom_region_pan_test (gpointer data)
 
 		if (roi.y1 >= screen_height - height) {
 			roi.y1 = screen_height - height;
-			reset_timing = TRUE;
 		}
 	} else {
 		if (pixel_direction == 0)
@@ -2255,10 +2377,8 @@ zoom_region_pan_test (gpointer data)
 	magnifier->priv->cursor_y = (roi.y2 + roi.y1) / 2;
 
 	/* Add one since in first loop we call zoom_region_process_updates */
-	if (counter > zoom_region->timing_iterations - 1) {
-		fprintf(stderr, "\n  Pan Rate                 = %d\n", zoom_region->timing_pan_rate);
+	if (counter > zoom_region->timing_iterations - 1)
 		zoom_region->exit_magnifier = TRUE;
-	}
 
 	zoom_regions = GNOME_Magnifier_Magnifier_getZoomRegions (
 		BONOBO_OBJREF (magnifier), &ev);
@@ -2270,11 +2390,14 @@ zoom_region_pan_test (gpointer data)
 
 	if (!finished_update) {
 		zoom_region_process_updates(zoom_region);
-		if (roi.y1 >= screen_height - height) {
+		if (roi.y1 == screen_height - height) {
 			finished_update = TRUE;
+			reset_timing = TRUE;
 		}
 	}
 
+    last_pixels_at_speed = pixels_at_speed;
+
 	return FALSE;
 }
 
@@ -2294,7 +2417,13 @@ impl_zoom_region_set_roi (PortableServer_Servant servant,
 	}
 
 	if (zoom_region->exit_magnifier) {
-		exit_report(zoom_region);
+		if (timing_test) {
+			fprintf(stderr, "\n### Timing Summary:\n\n");
+			if (zoom_region->timing_pan_rate)
+				fprintf(stderr, "  Pan Rate                 = %d\n", zoom_region->timing_pan_rate);
+			timing_report(zoom_region);
+		}
+		exit(0);
 	}
 
 	/*
@@ -2416,11 +2545,11 @@ impl_zoom_region_mark_dirty (PortableServer_Servant servant,
 {
 	ZoomRegion *zoom_region =
 		ZOOM_REGION (bonobo_object_from_servant (servant));
-	DEBUG_RECT ("mark dirty", zoom_region_source_rect_from_view_bounds (
+	DEBUG_RECT ("mark dirty", zoom_region_source_rect_clipto_view_bounds (
 			    zoom_region, roi_dirty) );
 	zoom_region_update_pointer (zoom_region, FALSE);
-	zoom_region_queue_update (zoom_region, zoom_region_source_rect_from_view_bounds
-				  (zoom_region, roi_dirty));
+	zoom_region_queue_update (zoom_region, zoom_region_clip_to_target
+				  (zoom_region, zoom_region_rect_from_bounds (zoom_region, roi_dirty)));
 }
 
 static GNOME_Magnifier_RectBounds
